/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 1999-2000 Michael H. Schimek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: mmx_q_fdct_inter.c,v 1.10 2000/06/08 18:53:34 michael Exp $ */

#include "ieee.h"
#include "../mpeg.h"
#include "../../misc.h"
#include "../../mmx.h"

/*
 *  MMX forward DCT fusioned with inter block quantization
 *
 *  Not really fast but more accurate, to address 9 bit samples
 *  and the the uniform quantization matrix which doesn't combine
 *  well with the AAN coeff LUT.
 */

static void init_lut(void) __attribute__ ((constructor));

static mmx_t c1, c2;
static short mmx_inter_l0[8][1];
static short mmx_inter_l2[8][2], mmx_inter_l6[8][2];
static short mmx_inter_l1[8][2], mmx_inter_l3[8][2], mmx_inter_l5[8][2], mmx_inter_l7[8][2];

static void
init_lut(void)
{
	int v, u;
	double mmx_inter_lut[8][8];

	for (v = 0; v < 8; v++) {
		for (u = 0; u < 8; u++) {
			double Cu, Cv, m;

    			Cu = (u == 0) ? 1.0 : (cos(u * IEEE_PI / 16.0) * sqrt(2.0));
			Cv = (v == 0) ? 1.0 : (cos(v * IEEE_PI / 16.0) * sqrt(2.0));

			if (v == 2 || v == 6) Cv = 1.0;

			m = 1.0 / (Cu * Cv * 8.0);

			if (u & 1) m *= C6;
			if (u == 0 || u == 4 || u == 7) m /= 4.0;
			if (u == 2 || u == 5) m /= 2.0;
			if (u == 6) m /= 8.0;

			mmx_inter_lut[v][u] = m;
		}
	}

	dump(mmx_inter_lut);

	c1 = MMXW(lroundn(C2 * S15), -lroundn(C6 * S15), lroundn(C2 * S15), -lroundn(C6 * S15));
	c2 = MMXW(lroundn(C6 * S15), +lroundn(C2 * S15), lroundn(C6 * S15), +lroundn(C2 * S15));

	for (u = 0; u < 8; u++) {
		mmx_inter_l0[u][0] = lroundn(mmx_inter_lut[0][u] * S19);
		mmx_inter_l2[u][0] = lroundn(mmx_inter_lut[0][u] * +(C2 + C6) * S18);
		mmx_inter_l2[u][1] = lroundn(mmx_inter_lut[0][u] * +(C2 - C6) * S18);
		mmx_inter_l6[u][0] = lroundn(mmx_inter_lut[0][u] * +(C2 - C6) * S18);
		mmx_inter_l6[u][1] = lroundn(mmx_inter_lut[0][u] * -(C2 + C6) * S18);
		mmx_inter_l1[u][1] = +(mmx_inter_l1[u][0] = lroundn(mmx_inter_lut[1][u] * S19));
		mmx_inter_l3[u][1] = -(mmx_inter_l3[u][0] = lroundn(mmx_inter_lut[3][u] * S19));
		mmx_inter_l5[u][1] = +(mmx_inter_l5[u][0] = lroundn(mmx_inter_lut[5][u] * S19));
		mmx_inter_l7[u][1] = -(mmx_inter_l7[u][0] = lroundn(mmx_inter_lut[7][u] * S17));
	}

	dump(mmx_inter_l0);
	dump(mmx_inter_l2); dump(mmx_inter_l6);
	dump(mmx_inter_l1); dump(mmx_inter_l3); dump(mmx_inter_l5); dump(mmx_inter_l7);
}

#define descale_quantize(lut, reg, bits)	\
	pmaddwd(*(mmx_t *) lut,&reg);		\
	paddd(MMXRD(1 << (bits - 1)),&reg);	\
	psrad(bits,&reg);			\
	pmaddwd(d,&reg);			\
	psrad(15,&reg);

void
mmx_q_fdct_inter(int q, short block[8][8])
{
	short t[8][8];
	int v, u, i, val, div;
	mmx_t d = MMXW(lroundn(32768.0 / q / 2.0), -32767, lroundn(32768.0 / q / 2.0), -32767);
	mmx_t mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7, mm8, mm9;

	for (u = 0; u < 8; u += 4) {
		pload(&block[7][u],&mm0); pload(&block[0][u],&mm7);	// 255 (Inter)
		pload(&block[6][u],&mm1); pload(&block[1][u],&mm6);
		pload(&block[5][u],&mm2); pload(&block[2][u],&mm5);
		pload(&block[4][u],&mm3); pload(&block[3][u],&mm4);

		paddw(mm7,&mm0); paddw(mm7,&mm7); psubw(mm0,&mm7);	// 510
		paddw(mm6,&mm1); paddw(mm6,&mm6); psubw(mm1,&mm6);
		paddw(mm5,&mm2); paddw(mm5,&mm5); psubw(mm2,&mm5);
		paddw(mm4,&mm3); paddw(mm4,&mm4); psubw(mm3,&mm4);

		/* even butterfly */

		paddw(mm0,&mm3); paddw(mm0,&mm0); psubw(mm3,&mm0);	// 1020
		paddw(mm1,&mm2); paddw(mm1,&mm1); psubw(mm2,&mm1);

		paddw(mm3,&mm2); paddw(mm2,&mm2);
		psllw(2,&mm3); psubw(mm2,&mm3);				// 8160

		paddw(mm0,&mm1); psllw(2,&mm0);
		psllw(3,&mm1); paddw(MMXRW(1),&mm1);
		pmulhw(MMXRW(lroundn(C4 * S15)),&mm1);
		paddw(mm0,&mm1); paddw(mm0,&mm0); psubw(mm1,&mm0);
									// [mm0' * (1 + sqrt(2))]
		paddw(MMXRW(2),&mm1); psraw(2,&mm1);
								
		pstore(mm0,&t[6][u]);
		pstore(mm1,&t[2][u]);
		pstore(mm2,&t[0][u]);
		pstore(mm3,&t[4][u]);

		/* odd butterfly */

		paddw(mm5,&mm4); paddw(mm6,&mm5); paddw(mm7,&mm6);	// 1020

		psllw(4,&mm5); psllw(5,&mm7);		
		pmulhw(MMXRW(lroundn(C4 / C6 * S14)),&mm5);
		pmulhw(MMXRW(lroundn(1.0 / C6 * S13)),&mm7);
		paddw(mm7,&mm5); paddw(mm7,&mm7); psubw(mm5,&mm7);
		
		movq(mm4,&mm2); psllw(4,&mm2);
		movq(mm6,&mm3); psllw(4,&mm3);

		pmulhw(MMXRW(lroundn((C2 / C6 - 1) * S14)),&mm2);
		pmulhw(MMXRW(lroundn((C2 / C6 - 1) * S14)),&mm3);
		// constant 23170, same as C4 << 15 above

		paddw(mm4,&mm6); psllw(3,&mm4); psllw(2,&mm6); psubw(mm6,&mm4);
		paddw(mm2,&mm4); paddw(mm3,&mm6);

		paddw(mm7,&mm4); paddw(mm7,&mm7); psubw(mm4,&mm7);
		paddw(mm5,&mm6); paddw(mm5,&mm5); psubw(mm6,&mm5);

		paddw(MMXRW(4),&mm6); psraw(3,&mm6);
		paddw(MMXRW(4),&mm7); psraw(3,&mm7);
		paddw(MMXRW(2),&mm4); psraw(2,&mm4);
		paddw(MMXRW(1),&mm5); psraw(1,&mm5);

		pstore(mm6,&t[1][u]);
		pstore(mm7,&t[3][u]);
    		pstore(mm4,&t[5][u]);
		pstore(mm5,&t[7][u]);
	}

	mirror(t);

	for (u = 0; u < 8; u += 4) {
		pload(&t[7][u],&mm0); pload(&t[0][u],&mm7);	
		pload(&t[6][u],&mm1); pload(&t[1][u],&mm6);
		pload(&t[5][u],&mm2); pload(&t[2][u],&mm5);
		pload(&t[4][u],&mm3); pload(&t[3][u],&mm4);

		paddw(mm7,&mm0); paddw(mm7,&mm7); psubw(mm0,&mm7);	
		paddw(mm6,&mm1); paddw(mm6,&mm6); psubw(mm1,&mm6);
		paddw(mm5,&mm2); paddw(mm5,&mm5); psubw(mm2,&mm5);
		paddw(mm4,&mm3); paddw(mm4,&mm4); psubw(mm3,&mm4);

		/* even */

		paddw(mm0,&mm3); paddw(mm0,&mm0); psubw(mm3,&mm0);
		paddw(mm1,&mm2); paddw(mm1,&mm1); psubw(mm2,&mm1);

		paddw(mm3,&mm2); paddw(mm3,&mm3); psubw(mm2,&mm3);

		pmulhw(*(mmx_t *) &mmx_inter_l0[u],&mm3); paddw(MMXRW(2),&mm3);
		pmulhw(*(mmx_t *) &mmx_inter_l0[u],&mm2); paddw(MMXRW(2),&mm2);

		movq(mm2,&mm8); movq(mm3,&mm9);
		punpcklwd(mm2,&mm2); psrad(18,&mm2); pmaddwd(d,&mm2); psrad(15,&mm2);
		punpcklwd(mm3,&mm3); psrad(18,&mm3); pmaddwd(d,&mm3); psrad(15,&mm3);
		punpckhwd(mm8,&mm8); psrad(18,&mm8); pmaddwd(d,&mm8); psrad(15,&mm8);		
		punpckhwd(mm9,&mm9); psrad(18,&mm9); pmaddwd(d,&mm9); psrad(15,&mm9);		
		packssdw(mm8,&mm2);
		packssdw(mm9,&mm3);

		pstore(mm2,&block[0][u]);
		pstore(mm3,&block[4][u]);

		movq(mm0,&mm2);
		punpckhwd(mm1,&mm2); movq(mm2,&mm3);
		punpcklwd(mm1,&mm0); movq(mm0,&mm1);

		descale_quantize(&mmx_inter_l2[u + 0], mm0, 17);
		descale_quantize(&mmx_inter_l2[u + 2], mm2, 17);
		descale_quantize(&mmx_inter_l6[u + 0], mm1, 17);
		descale_quantize(&mmx_inter_l6[u + 2], mm3, 17);

		packssdw(mm2,&mm0);
		packssdw(mm3,&mm1);

		pstore(mm0,&block[2][u]);
		pstore(mm1,&block[6][u]);

		/* odd */

		paddw(mm5,&mm4); paddw(mm6,&mm5); paddw(mm7,&mm6);

		paddw(mm5,&mm5); paddw(MMXRW(1),&mm5);
		pmulhw(MMXRW(lroundn(C4 * S15)),&mm5);
		paddw(mm7,&mm5); paddw(mm7,&mm7); psubw(mm5,&mm7);

		movq(mm5,&mm0);
		movq(mm7,&mm1);
	
		paddw(mm4,&mm4); paddw(MMXRW(1),&mm4);
		paddw(mm6,&mm6); paddw(MMXRW(1),&mm6);

		movq(mm4,&mm2);
		punpckhwd(mm6,&mm2); movq(mm2,&mm3);
		punpcklwd(mm6,&mm4); movq(mm4,&mm6);

		pmaddwd(c1,&mm4); psrad(16,&mm4);
		pmaddwd(c2,&mm6); psrad(16,&mm6);
		pmaddwd(c1,&mm2); psrad(16,&mm2);
		pmaddwd(c2,&mm3); psrad(16,&mm3);

		packssdw(mm2,&mm4);
		packssdw(mm3,&mm6);

		punpcklwd(mm6,&mm0); movq(mm0,&mm3);
		punpcklwd(mm4,&mm1); movq(mm1,&mm2);

		descale_quantize(&mmx_inter_l1[u + 0], mm0, 18);
		descale_quantize(&mmx_inter_l3[u + 0], mm1, 18);
		descale_quantize(&mmx_inter_l5[u + 0], mm2, 18);
		descale_quantize(&mmx_inter_l7[u + 0], mm3, 16);

		punpckhwd(mm6,&mm5); movq(mm5,&mm6);
		punpckhwd(mm4,&mm7); movq(mm7,&mm4);

		descale_quantize(&mmx_inter_l1[u + 2], mm5, 18);
		descale_quantize(&mmx_inter_l3[u + 2], mm7, 18);
		descale_quantize(&mmx_inter_l5[u + 2], mm4, 18);
		descale_quantize(&mmx_inter_l7[u + 2], mm6, 16);

		packssdw(mm5,&mm0);
		packssdw(mm7,&mm1);
		packssdw(mm4,&mm2);
		packssdw(mm6,&mm3);

		pstore(mm0,&block[1][u]);
		pstore(mm1,&block[3][u]);
		pstore(mm2,&block[5][u]);
		pstore(mm3,&block[7][u]);
	}

	for (i = 0; i < 64; i++) {
		// val = block[0][i];
		// div = lroundn(65536.0 / q / 2.0);
		// val = val * div + (val < 0 ? 65535 : 0);
		// block[0][i] = saturate(val >> 16, -255, +255);

		block[0][i] = saturate(block[0][i], -255, +255);
		// saturation redundant if q > 1
	}

	mirror(block);
}
