/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 1999-2000 Michael H. Schimek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: mmx_q_fdct_intra.c,v 1.17 2000/06/08 18:53:34 michael Exp $ */

#include "ieee.h"
#include "../mpeg.h"
#include "../../misc.h"
#include "../../mmx.h"

/*
 *  MMX forward DCT fusioned with intra block quantization
 */

extern char ltp[32];
extern char lts[32];

static short mmx_q_fdct_intra_q_lut[8][8][8];
static char mmx_q_fdct_intra_sh[32];

static void init_lut(void) __attribute__ ((constructor));

static void
init_lut(void)
{
	int q, v, u, sh, max;
        double Cu, Cv, m;
	int shq[8];

	for (q = 0; q < 8; q++) {
		for (sh = max = 0; max < 16384; sh++)
			for (v = max = 0; v < 8; v++)
				for (u = 0; u < 8; u++) {
					Cu = (u == 0) ? 1 : (cos(u * IEEE_PI / 16.0) * sqrt(2.0));
					Cv = (v == 0) ? 1 : (cos(v * IEEE_PI / 16.0) * sqrt(2.0));

					m = 1.0 / (Cu * Cv * 8.0);

					if (u == 0 || u == 4) m *= 0.125;
					if (u == 2 || u == 6) m *= 0.25;
					if (u & 1) m *= C6 * 0.5;
					if (v == 0 || v == 4) m *= 2;
					if (v == 2 || v == 6) m *= 4;
					if (v & 1) m *= C6 * 8;

					mmx_q_fdct_intra_q_lut[q][v][u] = lroundn(
						m
						* 8
						/ default_intra_quant_matrix[v][u]
						/ (2 * q + 1)
						* (double)(1 << sh));

					if (mmx_q_fdct_intra_q_lut[q][v][u] > max)
						max = mmx_q_fdct_intra_q_lut[q][v][u];
				}

		shq[q] = sh;

		// dump(mmx_q_fdct_intra_q_lut[q]);
	}

	for (q = 1; q < 32; q++) {
		int ltsi = lts[q], ltpi = ltp[q];
		mmx_q_fdct_intra_sh[q] = shq[ltpi] + ltsi - 17;
	}
}

void
mmx_q_fdct_intra(int q, short block[8][8])
{
	short t[8][8];
	mmx_t mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7;
	int sh, rnd, ltpi = ltp[q];
	int val, u;

	sh = mmx_q_fdct_intra_sh[q];
	rnd = 1 << (sh - 1);

	for (u = 0; u < 8; u += 4) {
		pload(&block[7][u],&mm0); pload(&block[0][u],&mm7);		// -128 ... +127 (Intra)
		pload(&block[6][u],&mm1); pload(&block[1][u],&mm6);
		pload(&block[5][u],&mm2); pload(&block[2][u],&mm5);
		pload(&block[4][u],&mm3); pload(&block[3][u],&mm4);

		paddw(mm7,&mm0); paddw(mm7,&mm7); psubw(mm0,&mm7);		// 256
		paddw(mm6,&mm1); paddw(mm6,&mm6); psubw(mm1,&mm6);
		paddw(mm5,&mm2); paddw(mm5,&mm5); psubw(mm2,&mm5);
		paddw(mm4,&mm3); paddw(mm4,&mm4); psubw(mm3,&mm4);

		/* even butterfly */

		paddw(mm0,&mm3); paddw(mm0,&mm0); psubw(mm3,&mm0);		// 512
		paddw(mm1,&mm2); paddw(mm1,&mm1); psubw(mm2,&mm1);

		paddw(mm3,&mm2); psllw(3,&mm2);
		psllw(4,&mm3); psubw(mm2,&mm3);					// 8192

		paddw(mm0,&mm1); psllw(2,&mm0);
		psllw(3,&mm1); pmulhw(MMXRW(lroundn(C4 * S15)),&mm1);
		paddw(mm0,&mm1); paddw(mm0,&mm0); psubw(mm1,&mm0);
						    // 4944 [4 * mm0' * (1 + sqrt(2))]

		pstore(mm0,&t[6][u]);
		pstore(mm1,&t[2][u]);
		pstore(mm2,&t[0][u]);
		pstore(mm3,&t[4][u]);

		/* odd butterfly */

		paddw(mm5,&mm4); paddw(mm6,&mm5); paddw(mm7,&mm6);		// 512

		psllw(3,&mm5); psllw(4,&mm7);					// 4096

		pmulhw(MMXRW(lroundn(C4 / C6 * S14)),&mm5);
		pmulhw(MMXRW(lroundn(1.0 / C6 * S13)),&mm7);

		movq(mm4,&mm2); psllw(3,&mm2);
		movq(mm6,&mm3); psllw(3,&mm3);

		pmulhw(MMXRW(lroundn((C2 / C6 - 1) * S14)),&mm2);
		pmulhw(MMXRW(lroundn((C2 / C6 - 1) * S14)),&mm3);
		// constant 23170, same as C4 << 15 above

		paddw(mm7,&mm5); paddw(mm7,&mm7); psubw(mm5,&mm7);		// 3230

		paddw(mm4,&mm6); psllw(2,&mm4); paddw(mm6,&mm6); psubw(mm6,&mm4);
		paddw(mm2,&mm4);			// 3496 [2 * mm4' * (2 + sqrt(2))]
		paddw(mm3,&mm6);

		paddw(mm7,&mm4); paddw(mm7,&mm7); psubw(mm4,&mm7);		// 6726
		paddw(mm5,&mm6); paddw(mm5,&mm5); psubw(mm6,&mm5);

    		pstore(mm4,&t[5][u]);
		pstore(mm5,&t[7][u]);
		pstore(mm6,&t[1][u]);
		pstore(mm7,&t[3][u]);
	}

	mirror(t);

	for (u = 0; u < 8; u += 4) {
		pload(&t[7][u],&mm0); pload(&t[0][u],&mm7);			// 8192
		pload(&t[6][u],&mm1); pload(&t[1][u],&mm6);
		pload(&t[5][u],&mm2); pload(&t[2][u],&mm5);
		pload(&t[4][u],&mm3); pload(&t[3][u],&mm4);

		paddw(MMXRW(1),&mm6); // rounding magic
		paddw(MMXRW(1),&mm5);
		paddw(MMXRW(1),&mm4);

		paddw(mm7,&mm0); paddw(mm7,&mm7); psubw(mm0,&mm7);		// 16384
		paddw(mm6,&mm1); paddw(mm6,&mm6); psubw(mm1,&mm6);
		paddw(mm5,&mm2); paddw(mm5,&mm5); psubw(mm2,&mm5);
		paddw(mm4,&mm3); paddw(mm4,&mm4); psubw(mm3,&mm4);

		/* even butterfly */

		paddw(mm0,&mm3); paddw(mm0,&mm0); psubw(mm3,&mm0);		// 32768
		paddw(mm1,&mm2); paddw(mm1,&mm1); psubw(mm2,&mm1);

		psraw(1,&mm2); psraw(1,&mm3);
		paddw(mm3,&mm2); paddw(mm3,&mm3); psubw(mm2,&mm3);		// 32768

		psraw(1,&mm0); psraw(1,&mm1);
		paddw(mm0,&mm1); pmulhw(MMXRW(lroundn(C4 * S15)),&mm1);
		psraw(1,&mm0);
		paddw(mm0,&mm1); paddw(mm0,&mm0); psubw(mm1,&mm0);		// 19777

		pstore(mm0,&block[6][u]);
		pstore(mm1,&block[2][u]);
		pstore(mm2,&block[0][u]);
		pstore(mm3,&block[4][u]);

		/* odd butterfly */

		paddw(mm5,&mm4);						// 32768
		paddw(mm6,&mm5);
		paddw(mm7,&mm6);

		pmulhw(MMXRW(lroundn(C4 / C6 * S13)),&mm5);
		pmulhw(MMXRW(lroundn(1.0 / C6 * S13)),&mm7);

		movq(mm4,&mm2); movq(mm6,&mm3);

		pmulhw(MMXRW(lroundn((C2 / C6 - 1) * S13)),&mm2);
		pmulhw(MMXRW(lroundn((C2 / C6 - 1) * S13)),&mm3);
		// same as C4 << 14

		psraw(3,&mm4); psraw(3,&mm6);
		paddw(mm4,&mm6); paddw(mm4,&mm4); psubw(mm6,&mm4);		// 8192
		paddw(mm2,&mm4); paddw(mm3,&mm6);				// 19777

		paddw(mm7,&mm5); paddw(mm7,&mm7); psubw(mm5,&mm7);		// 12920

		paddw(mm7,&mm4); paddw(mm7,&mm7); psubw(mm4,&mm7);		// 32697
		paddw(mm5,&mm6); paddw(mm5,&mm5); psubw(mm6,&mm5);

		pstore(mm4,&block[5][u]);
		pstore(mm5,&block[7][u]);
		pstore(mm6,&block[1][u]);
		pstore(mm7,&block[3][u]);
	}

	mirror(block);

	block[0][0] = (block[0][0] + 128) >> 8; // saturation redundant

	for (u = 1; u < 64; u++) {
		val = (((block[0][u] * mmx_q_fdct_intra_q_lut[ltpi][0][u]) >> 16) + rnd) >> sh;
		block[0][u] = val;
		// block[0][u] = saturate(val, -255, +255);
		// saturation redundant if q > 1
	}
}
