/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 1999-2000 Michael H. Schimek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: vlc.c,v 1.9 2000/06/08 18:53:34 michael Exp $ */

#include <assert.h>
#include <limits.h>
#include "../bstream.h"
#include "../log.h"
#include "mpeg.h"
#include "vlc.h"

int			dc_dct_pred[2][3];

VLC2			coded_block_pattern[64];
VLC2			macroblock_address_increment[33];

VLC4
macroblock_type_b_nomc_quant[4] =
{
	{ 0x0020, 11, 0 },	// '0000 01 xxxxx' (B Intra, Quant)
	{ 0x0183, 13, 2 },	// '0000 11 xxxxx 11' (B Fwd Coded, Quant, MV (0, 0))
	{ 0x0103, 13, 2 },	// '0000 10 xxxxx 11' (B Bwd Coded, Quant, MV (0, 0))
	{ 0x040F, 14, 4 }	// '00 01 0 xxxxx 11 11' (B Interp Coded, Quant, FMV (0, 0), BMV (0, 0))
};

VLC2
macroblock_type_b_nomc[4] =
{
	{ 0x0F, 7 },		// '0001 1  11' (B Intra, MV (0, 0))
	{ 0x0F, 6 },		// '0011  11' (B Fwd, Coded, MV (0, 0))
	{ 0x0F, 5 },		// '011  11' (B Bwd, Coded, MV (0, 0))
	{ 0x3F, 6 },		// '11  11 11' (B Interp, Coded, FMV (0, 0), BMV (0, 0))
};

VLC2
macroblock_type_b_notc[4] =
{
	{ 0, 0 },
	{ 0x0B, 6 },		// '1  0010  11' (Addr +1, B Fwd, Not Coded, MV (0, 0))
	{ 0x0B, 5 },		// '1  010  11' (Addr +1, B Bwd, Not Coded, MV (0, 0))
	{ 0x2F, 6 },		// '1  10  11 11' (Addr +1, B Interp, Not Coded, FMV (0, 0), BMV (0, 0))
};

unsigned char		iscan[8][8];
VLC8			dc_vlc_intra[5][12];
VLC2			ac_vlc_zero[176];
VLC2			ac_vlc_one[176];

extern short		mblock[7][6][8][8];
extern struct bs_rec	video_out;
extern const char 	cbp_order[6];

void
vlc_init(void)
{
	int i, j;
	unsigned int code;
	int dct_dc_size;
	int run, level, length;

	/* Variable length codes for macroblock address increment */

	for (i = 0; i < 33; i++) {
		macroblock_address_increment[i].length = vlc(macroblock_address_increment_vlc[i], &code);
		macroblock_address_increment[i].code = code;
		assert(code <= UCHAR_MAX);
	}

	/* Variable length codes for coded block pattern */

	for (i = 0; i < 64; i++) {
		int j, k;

		for (j = k = 0; k < 6; k++)
			if (i & (1 << k))
				j |= 0x20 >> cbp_order[k]; // (5 - k)

		coded_block_pattern[i].length = vlc(coded_block_pattern_vlc[j], &code);
		coded_block_pattern[i].code = code;
		assert(code <= UCHAR_MAX);
	}

	/* Variable length codes for intra DC coefficient */

	for (dct_dc_size = 0; dct_dc_size < 12; dct_dc_size++) {
		// Intra DC luma VLC
		dc_vlc_intra[0][dct_dc_size].length = vlc(dct_dc_size_luma_vlc[dct_dc_size], &code) + dct_dc_size;
		dc_vlc_intra[0][dct_dc_size].code = code << dct_dc_size;

		// Intra DC luma VLC with EOB ('10' table B-14) of previous block
		dc_vlc_intra[1][dct_dc_size].length = vlc(dct_dc_size_luma_vlc[dct_dc_size], &code) + dct_dc_size + 2;
		dc_vlc_intra[1][dct_dc_size].code = ((0x2 << vlc(dct_dc_size_luma_vlc[dct_dc_size], &code)) | code) << dct_dc_size;

		// Intra DC chroma VLC with EOB of previous block
		dc_vlc_intra[2][dct_dc_size].length = vlc(dct_dc_size_chroma_vlc[dct_dc_size], &code) + dct_dc_size + 2;
		dc_vlc_intra[2][dct_dc_size].code = ((0x2 << vlc(dct_dc_size_chroma_vlc[dct_dc_size], &code)) | code) << dct_dc_size;

		// Intra DC luma VLC with EOB ('0110' table B-15) of previous block
		dc_vlc_intra[3][dct_dc_size].length = vlc(dct_dc_size_luma_vlc[dct_dc_size], &code) + dct_dc_size + 4;
		dc_vlc_intra[3][dct_dc_size].code = ((0x6 << vlc(dct_dc_size_luma_vlc[dct_dc_size], &code)) | code) << dct_dc_size;

		// Intra DC chroma VLC with EOB of previous block
		dc_vlc_intra[4][dct_dc_size].length = vlc(dct_dc_size_chroma_vlc[dct_dc_size], &code) + dct_dc_size + 4;
		dc_vlc_intra[4][dct_dc_size].code = ((0x6 << vlc(dct_dc_size_chroma_vlc[dct_dc_size], &code)) | code) << dct_dc_size;
	}

	/* Variable length codes for AC coefficients (table B-14) */

	for (i = run = 0; run < 64; run++) {
		assert(i <= sizeof(ac_vlc_zero) / sizeof(ac_vlc_zero[0]));

		ac_vlc_zero[j = i++].code = run;

		for (level = 1; (length = dct_coeff_vlc(0, run, level, &code)) > 0; level++, i++) {
			assert(i < sizeof(ac_vlc_zero) / sizeof(ac_vlc_zero[0]));
			assert((code << 1) <= UCHAR_MAX);

			ac_vlc_zero[i].length = length + 1;
			ac_vlc_zero[i].code = code << 1; // sign 0
		}

		ac_vlc_zero[j].length = i - j;
	}

	/* Variable length codes for AC coefficients (table B-15) */

	for (i = run = 0; run < 64; run++) {
		assert(i <= sizeof(ac_vlc_one) / sizeof(ac_vlc_one[0]));

		ac_vlc_one[j = i++].code = run;

		for (level = 1; (length = dct_coeff_vlc(1, run, level, &code)) > 0; level++, i++) {
			assert(i < sizeof(ac_vlc_one) / sizeof(ac_vlc_one[0]));
			assert((code << 0) <= UCHAR_MAX);

			ac_vlc_one[i].length = length + 1;
			ac_vlc_one[i].code = code << 0; // no sign (would need 9 bits)
		}

		ac_vlc_zero[j].length = i - j;
	}

	/*
	 *  Forward zig-zag scanning pattern
	 *  
	 *  Reversed because the vlc routines count down, and
	 *  mirrored to remove mirror operations in FDCT and
	 *  IDCT.
	 */
	for (i = 0; i < 64; i++) {
		iscan[0][63 - scan[0][0][i]] = (i & 7) * 8 + (i >> 3);
	}
}

/* Reference */

int
mpeg1_encode_intra(void)
{
	int v;

	int
	encode_block(short block[8][8], int *dc_pred, VLC8 *dc_vlc)
	{
		/* DC coefficient */

		{
			register int val = block[0][0] - *dc_pred, size;
			
			/*
			 *  Find first set bit, starting at msb with 0 -> 0.
			 */
			asm("
				bsrl		%1,%0
				jnz		1f
				movl		$-1,%0
1:		    		incl		%0
			" : "=r" (size) : "r" (abs(val)));

			if (val < 0) {
				val--;
				val ^= (-1 << size);
			}

			bputl(&video_out, dc_vlc[size].code | val, dc_vlc[size].length);

			*dc_pred = block[0][0];
		}

		/* AC coefficients */

		{
			VLC2 *p = ac_vlc_zero;
			int i;

			for (i = 1; i < 64; i++) {
	    			int ulevel, slevel = block[0][iscan[0][63 - i]];

				if (slevel) {
					ulevel = abs(slevel);

		    			if (ulevel < (int) p->length) {
						p += ulevel;
						bputl(&video_out, p->code | ((slevel >> 31) & 1), p->length);
					} else {
		    				int len;

		    				if (slevel > 127) {
							if (slevel > 255)
								return 1;
							/* %000001 escape, 6 bit run, %00000000, slevel & 0xFF */
							slevel = 0x0400000 | (p->code << 16) | (slevel & 0xFF);
							len = 28;
						} else if (slevel < -127) {
							if (slevel < -255)
								return 1;
							/* %000001 escape, 6 bit run, %10000000, slevel (sic) & 0xFF */
							slevel = 0x0408000 | (p->code << 16) | (slevel & 0xFF);
							len = 28;
						} else {
							/* %000001 escape, 6 bit run, slevel & 0xFF */
							slevel = (1 << 14) | (p->code << 8) | (slevel & 0xFF);
							len = 20;
						}

						bputl(&video_out, slevel, len);
					}

					p = ac_vlc_zero; // run = 0
				} else
					p += p->length; // run++
			}
		}

		return 0;
	}

	dc_dct_pred[1][0] = dc_dct_pred[0][0];
	dc_dct_pred[1][1] = dc_dct_pred[0][1];
	dc_dct_pred[1][2] = dc_dct_pred[0][2];

	v  = encode_block(mblock[1][0], &dc_dct_pred[0][0], dc_vlc_intra[0]);
	v |= encode_block(mblock[1][2], &dc_dct_pred[0][0], dc_vlc_intra[1]);
	v |= encode_block(mblock[1][1], &dc_dct_pred[0][0], dc_vlc_intra[1]);
	v |= encode_block(mblock[1][3], &dc_dct_pred[0][0], dc_vlc_intra[1]);
	v |= encode_block(mblock[1][4], &dc_dct_pred[0][1], dc_vlc_intra[2]);
	v |= encode_block(mblock[1][5], &dc_dct_pred[0][2], dc_vlc_intra[2]);

	bputl(&video_out, 0x2, 2); // EOB '10' (ISO 13818-2 table B-14)

	/*
	 *  Saturation is rarely needed, so the forward quantisation code
	 *  skips the step. This routine detects excursions in uncritical
	 *  path and reports but saturates because saturation often causes
	 *  a visibly annoying reconstruction error.
	 */
	if (v) {
		dc_dct_pred[0][0] = dc_dct_pred[1][0];
		dc_dct_pred[0][1] = dc_dct_pred[1][1];
		dc_dct_pred[0][2] = dc_dct_pred[1][2];
	}

	return v;
}

int
mpeg1_encode_inter(short iblock[6][8][8], unsigned int cbp)
{
	int v = 0;

	int
	encode_block(short block[8][8])
	{
		VLC2 *p = ac_vlc_zero; // ISO 13818-2 table B-14
    		int i = 1, len, ulevel, slevel;

		/* DC coefficient */

		ulevel = abs(slevel = block[0][0]);

		if (ulevel == 1) {
			bputl(&video_out, 0x2 | ((slevel >> 31) & 1), 2);
		} else
			i = 0;

		/* AC coefficients */

		while (i < 64) {
	    		if ((slevel = block[0][iscan[0][63 - i]])) {
				ulevel = abs(slevel);

		    		if (ulevel < (int) p->length) {
					p += ulevel;
					bputl(&video_out, p->code | ((slevel >> 31) & 1), p->length);
				} else {
		    			if (slevel > 127) {
						if (slevel > 255)
							return 1;
						/* %000001 escape, 6 bit run, %00000000, slevel & 0xFF */
						slevel = 0x0400000 | (p->code << 16) | (slevel & 0xFF);
						len = 28;
					} else if (slevel < -127) {
						if (slevel < -255)
							return 1;
						/* %000001 escape, 6 bit run, %10000000, slevel (sic) & 0xFF */
						slevel = 0x0408000 | (p->code << 16) | (slevel & 0xFF);
						len = 28;
					} else {
						/* %000001 escape, 6 bit run, slevel & 0xFF */
						slevel = (1 << 14) | (p->code << 8) | (slevel & 0xFF);
						len = 20;
					}

					bputl(&video_out, slevel, len);
				}

				p = ac_vlc_zero; // run = 0
			} else
				p += p->length; // run++
			i++;
		}

		bputl(&video_out, 0x2, 2);
		return 0;
	}

	// watch cbp_order
	if (cbp & (1 << 5)) v  = encode_block(iblock[0]);
	if (cbp & (1 << 3)) v |= encode_block(iblock[2]);
	if (cbp & (1 << 4)) v |= encode_block(iblock[1]);
	if (cbp & (1 << 2)) v |= encode_block(iblock[3]);
	if (cbp & (1 << 1)) v |= encode_block(iblock[4]);
	if (cbp & (1 << 0)) v |= encode_block(iblock[5]);

	return v;
}

// XXX add saturation cf. idct

void
mpeg2_encode_intra(void)
{
	void
	encode_block(short block[8][8], int *dc_pred, VLC8 *dc_vlc)
	{
		/* DC coefficient */

		{
			register int val = block[0][0] - *dc_pred, size;
			
			/*
			 *  Find first set bit, starting at msb with 0 -> 0.
			 */
			asm("
				bsrl		%1,%0
				jnz		1f
				movl		$-1,%0
1:		    		incl		%0
			" : "=r" (size) : "r" (abs(val)));

			if (val < 0) {
				val--;
				val ^= (-1 << size);
			}

			bputl(&video_out, dc_vlc[size].code | val, dc_vlc[size].length);

			*dc_pred = block[0][0];
		}

		/* AC coefficients */

		{
			VLC2 *p = ac_vlc_one;
			int i;

			for (i = 1; i < 64; i++) {
	    			int ulevel, slevel = block[0][iscan[0][63 - i]];

				if (slevel) {
					ulevel = abs(slevel);

		    			if (ulevel < (int) p->length) {
						p += ulevel;
						bputl(&video_out, (p->code << 1) | (slevel < 0), p->length);
					} else
						bputl(&video_out, (1 << 18) | (p->code << 12) | slevel, 24);
						// %000001 escape, 6 bit run, 12 bit slevel

					p = ac_vlc_one; // run = 0
				} else
					p += p->length; // run++
			}
		}
	}

	encode_block(mblock[1][0], &dc_dct_pred[0][0], dc_vlc_intra[0]);
	encode_block(mblock[1][2], &dc_dct_pred[0][0], dc_vlc_intra[3]);
	encode_block(mblock[1][1], &dc_dct_pred[0][0], dc_vlc_intra[3]);
	encode_block(mblock[1][3], &dc_dct_pred[0][0], dc_vlc_intra[3]);
	encode_block(mblock[1][4], &dc_dct_pred[0][1], dc_vlc_intra[4]);
	encode_block(mblock[1][5], &dc_dct_pred[0][2], dc_vlc_intra[4]);

	bputl(&video_out, 0x6, 4); // EOB '0110' (ISO 13818-2 table B-15)
}

void
mpeg2_encode_inter(short iblock[6][8][8], unsigned int cbp)
{
	void
	encode_block(short block[8][8])
	{
		VLC2 *p = ac_vlc_zero; // ISO 13818-2 table B-14
    		int i = 1, ulevel, slevel;

		/* DC coefficient */

		ulevel = abs(slevel = block[0][0]);

		if (ulevel == 1) {
			bputl(&video_out, 0x2 | ((slevel >> 31) & 1), 2);
		} else
			i = 0;

		/* AC coefficients */

		while (i < 64) {
	    		if ((slevel = block[0][iscan[0][63 - i]])) {
				ulevel = abs(slevel);

		    		if (ulevel < (int) p->length) {
					p += ulevel;
					bputl(&video_out, p->code | ((slevel >> 31) & 1), p->length);
				} else
					bputl(&video_out, (1 << 18) | (p->code << 12) | slevel, 24);
					// %000001 escape, 6 bit run, 12 bit slevel

				p = ac_vlc_zero; // run = 0
			} else
				p += p->length; // run++
			i++;
		}

		bputl(&video_out, 0x2, 2);
	}

	// watch cbp_order
	if (cbp & (1 << 5)) encode_block(iblock[0]);
	if (cbp & (1 << 3)) encode_block(iblock[2]);
	if (cbp & (1 << 4)) encode_block(iblock[1]);
	if (cbp & (1 << 2)) encode_block(iblock[3]);
	if (cbp & (1 << 1)) encode_block(iblock[4]);
	if (cbp & (1 << 0)) encode_block(iblock[5]);
}
